// Simple text editor

#include "xgc_editor.h"

// Document vars
static char doc_filename[MAX_FILENAME_LEN+1];
static char doc_text[MAX_DOC_LEN+1];
static int doc_final_char;
static int doc_file_open;
static int cursor;	// Cursor is position in the array

// Static consts
static const int start_line=3;
static const int end_line=MAX_LINES-3;
static const int cmd_line=MAX_LINES-1;
static const int status_line=1;
static const int cmd_status=MAX_LINES-2;
static const int max_chars=MAX_CHARS_LINE;

// Commands
static char cmd_text[MAX_EDIT_CMD_LEN + 1];
static int cmd_final_char=0;
static int cmd_alert=0;
static int cmd_incmd=0;

// Program running?
static int running=1;

// Main editor
void editor(chanend from_services, char file[]) {
	int ch;
	
	// Clear screen
	services_graphics_cs(from_services);
	
	// Clear everything
	doc_clearstate();
	
	// If we were supplied with a filename then reuse the existing functions to load it
	if (safestrlen(file)>0) {
		// Assemble a command string
		safestrcpy(cmd_text, "load ");
		safestrcat(cmd_text, file);
		
		// Add the last character
		cmd_final_char = safestrlen(cmd_text);
		
		command_load(from_services);
	}
	
	// Initialise a blank workspace
	update_cmd_area(from_services);
	update_status(from_services);
	update_text_area(from_services);
	
	// By default the user can insert text, press ESC to enter command
	while (running) {
		// Grab a key
		from_services <: AK_KEY_REQ;
		from_services :> ch;
		
		// Clear screen
		services_graphics_cs(from_services);
		
		// If key is ESC, enter/exit command mode
		if (ch==KEY_ESC) {
			// If exiting the command clear command line
			if (cmd_incmd) {
				// Wipe array
				cmd_text[0]='\0';
				cmd_final_char=0;
				// Reset
				cmd_incmd=0;
				cmd_alert=0;
			} else {
				// Start
				cmd_incmd=1;
			}
		}
		// Otherwise, pass either through to command mode or insert
		else if (cmd_incmd) {
			// Add text to command mode and/or execute command
			if (ch==KEY_ENTER) {
				command_execute(from_services);
			} else {
				if (!((ch==KEY_DEL)||(ch==KEY_UP)||(ch==KEY_DOWN)||(ch==KEY_LEFT)||(ch==KEY_RIGHT))) {
					command_add_char(ch);
				}
			}
		} else {
			// Operate on the next key
			
			if ((ch==KEY_UP)||(ch==KEY_DOWN)||(ch==KEY_LEFT)||(ch==KEY_RIGHT)) {
				// If it's an arrow key then move the cursor
				move_cursor(ch);
			} else if (ch==KEY_BACK) {
				// Delete a character
				del_backchar();
			} else if (ch==KEY_DEL) {
				// Move cursor forward and do backspace
				if (cursor!=doc_final_char) {
					cursor++;
					del_backchar();
				}
			} else if (ch==KEY_TAB) {
				// Do a tab, just insert 4 spaces
				insert_char(' ');
				insert_char(' ');
				insert_char(' ');
				insert_char(' ');
			} else if (ch==KEY_ENTER) {
				// Return
				insert_char('\n');
		 	} else {
				// Otherwise insert the char insert_char
				insert_char((char)ch);
			}
		}
		
		// Redraw all
		update_cmd_area(from_services);
		update_status(from_services);
		update_text_area(from_services);
	}
}

// Delete a character (backspace)
void del_backchar(void) {
	// Backspace shift all characters left
	int i;
	
	if (cursor==0) return;
	
	for (i=cursor;i<=doc_final_char;i++) {
		doc_text[i-1] = doc_text[i];
	}
	
	cursor--;
	doc_final_char--;
}

// Insert a character
void insert_char(char ch) {
	// Inserting a character is just moving everything in the array up one
	int i;
	
	// Block if the array is full
	if (doc_final_char==MAX_DOC_LEN) return;
	
	// If cursor is at the end, just insert
	if (cursor<doc_final_char) {
		// Start from the back and work until the cursor position, shifting each one forward
		for (i=doc_final_char; i>=cursor; i--) {
			doc_text[i+1] = doc_text[i];
		}
		
		// Add char
		doc_text[i+1] = ch;
		cursor++;
		doc_final_char++;
	} else {
		// Add char
		doc_text[doc_final_char] = ch;
		cursor++;
		
		// Restore end
		doc_final_char++;
		doc_text[doc_final_char] = '\0';
	}
}

// Utility function to give the number of used blocks on a line
int pos_in_line(int cur) {
	int pos=0;
	int i;
	
	// So we can work backwards until we meet a NL or start
	for (i=cur; i>=0; i--) {
		if (doc_text[i]=='\n') break;
		
		pos++;
	}
	
	if (i==0) pos++;
	
	// Return
	return pos;
}

// Move the cursor
void move_cursor(int ch) {
	int i;
	int finished=0;
	
	switch (ch) {
		case KEY_UP:
			// To go up we go -max_chars or to the next new line, whichever is smaller
			for (i=cursor-1; i>cursor-max_chars; i--) {
				if (i<=0) break;
				
				if (doc_text[i]=='\n') {
					cursor=i;
					finished=1;
					break;
				}
				
			}
			
			if (!finished) {
				if ((cursor - max_chars)>=0) cursor -= max_chars;
			}

			break;
			
		case KEY_DOWN:
			// To go up we go -max_chars or to the next new line, whichever is smaller
			for (i=cursor+1; i<=doc_final_char; i++) {
				if (doc_text[i]=='\n') {
					cursor=i;
					finished=1;
					break;
				}
				
			}
			
			if (!finished) {
				if ((cursor + max_chars)<=doc_final_char) cursor += max_chars; else cursor = doc_final_char;
			} 
			break;
		case KEY_LEFT:
			if (cursor>0) {
				cursor--;
			}
			break;
		case KEY_RIGHT:
			if (cursor<doc_final_char) {
				cursor++;
			}
			break;
	}
}

// Update the text area
void update_text_area(chanend from_services) {
	// 40 chars per line, end_line-start_line number of lines (eg 22-3=19) per screen
	// Each refresh the screen is cleared
	// Redraw the text based on the contents of the position of the cursor
	int tmp;
	int i, j;
	int lines_per_screen = end_line-start_line;
	int finished=0;
	int start_char=0;
	
	// Which part of the text are we in?
	tmp = 0;
	
	// Need to do a lookahead to find where the cursor is
	// Note we only find the start, we don't draw the whole screen
	while (!finished) {
		// Set start
		start_char=tmp;
		
		// Each line
		for (j=0; j<lines_per_screen; j++) {
	
			// Loop chars
			for (i=0; i<max_chars; i++) {
				
				if (tmp==cursor) {
					// Set the start was the previous screenful
					finished=1;
					break;
				} 
				
				if (doc_text[tmp]=='\n') {
					// Break, going to the next line
					tmp++;
					break;
				} else tmp++;
			}
		}
		
		if (finished==1) break;
	}
	
	// Reset tmp
	tmp=start_char;
	finished=0;
	
	// Render screenfulls of text until we hit the cursor
	while (!finished) {
		
		// Each line
		for (j=0; j<lines_per_screen; j++) {
			
			// Set position
			services_graphics_set_line(from_services, start_line+j);
			services_graphics_set_char(from_services, 0);
			
			// Loop chars
			for (i=0; i<max_chars; i++) {
	
				// Draw the characters
				if (tmp==cursor) {
					services_graphics_add_char(from_services, '_');
					finished=1;
				} 
			
				// Stop
				if (doc_text[tmp]=='\0') {
					i=max_chars;
					j=lines_per_screen;
					break;
				}
			
				// Check for NL
				if (doc_text[tmp]=='\n') {
					// Break, going to the next line
					tmp++;
					break;
				} else {
					services_graphics_add_char(from_services, doc_text[tmp]);
					tmp++;
				}
				
			}
		}
	}
}

// Add some text to the command mode
void command_add_char(int ch) {
	// If backspace
	if (ch==KEY_BACK) {
		if (cmd_final_char!=0) {
			cmd_final_char--;
			cmd_text[cmd_final_char] = '\0';
		}
	} else if (safestrlen(cmd_text)<MAX_EDIT_CMD_LEN) {
		cmd_text[cmd_final_char] = (char)ch;
		cmd_final_char++;
		cmd_text[cmd_final_char] = '\0';
	}
}


// Update the status line
void update_status(chanend from_services) {
	int i=0;
	char txt[5];
	
	// Set to this line
	services_graphics_set_line(from_services, status_line);

	// Print the filename
	services_graphics_set_char(from_services, 1);
	if(!doc_file_open) {
		services_graphics_add_text(from_services, "NEWDOC.NEW");
	} else {
		services_graphics_add_text(from_services, doc_filename);
	}
	 
	// Print the percent
	if (doc_final_char!=0) i = (cursor*100/doc_final_char); else i=0;
	itoa(i, txt);

	services_graphics_set_char(from_services, (max_chars/8)*4);
	services_graphics_add_char(from_services, '(');
	services_graphics_add_text(from_services, txt);
	services_graphics_add_text(from_services, "%)");
	
	// Used chars
	services_graphics_set_char(from_services, (max_chars/8)*5);
	itoa(doc_final_char, txt);
	services_graphics_add_text(from_services, txt);
	
	// In insert or command mode?
	services_graphics_set_char(from_services, (max_chars/8)*7);
	if (cmd_incmd) {
		services_graphics_add_text(from_services, "CMD");
	} else {
		services_graphics_add_text(from_services, "INS");
	}
}

// Update the cmd area
void update_cmd_area(chanend from_services) {
	// Set to this line
	services_graphics_set_line(from_services, cmd_line);
	services_graphics_set_char(from_services, 2);
	
	// If not in command, line is blank
	if (cmd_incmd) {
		// Show command
		services_graphics_add_char(from_services, '>');
		
		// Write command array
		services_graphics_set_char(from_services, 3);
		services_graphics_add_text(from_services, cmd_text);
		
		// If no text, no !
		if (safestrlen(cmd_text)==0) cmd_alert=0;
		
		// Alert is set
		if (cmd_alert) {
			services_graphics_set_line(from_services, cmd_line);
			services_graphics_set_char(from_services, 1);
			services_graphics_add_char(from_services, '!');
		}
	}
}

// Execute the specified command
void command_execute(chanend from_services) {
	// If the command is valid, execute, otherwise display an "!"
	// Set to this line
	if (safestrncmp(cmd_text, "save", 4)==0) command_save(from_services);
	else if (safestrncmp(cmd_text, "load", 4)==0) command_load(from_services);
	else if (safestrncmp(cmd_text, "new", 3)==0) command_new(from_services);
	else if (safestrncmp(cmd_text, "quit", 4)==0) running=0;
	else cmd_alert = 1;
}

// Load file
void command_load(chanend from_services) {
	char filename[MAX_FILENAME_LEN+1];
	int rtn;
	
	// Need to validate there's enough space in the filename array
	if (safestrlen(cmd_text)>MAX_FILENAME_LEN+5) {
		command_message(from_services, "Filename too long");
		cmd_alert=1;
		return;
	}
		
	// See if there's anything
	substr(cmd_text, filename, 5, cmd_final_char, 1);
	
	// Validate
	if (validate_filename(from_services, filename)!=0) return;
	
	// Copy to filename
	safestrcpy(doc_filename, filename);
	
	// If file doesn't exist then create it
	if (services_sdcard_exists(from_services, filename)!=0) {
		command_save(from_services);
		command_message(from_services, "New file created");
		return;
	}
	
	// Try to get
	rtn = services_sdcard_load(from_services, doc_filename, doc_text, MAX_DOC_LEN);

	switch (rtn) {
		case 1:
			command_message(from_services, "SD card not inserted");
			cmd_alert=1;
			break;
		case 2:
			command_message(from_services, "Unable to access file");
			cmd_alert=1;
			break;
		case 3:
			command_message(from_services, "Unable to read entire file ");
			cmd_alert=1;
			break;
		case 4:
			command_message(from_services, "Filename too long");
			cmd_alert=1;
			break;
		default:
			command_message(from_services, "OK!");
			
			// Command
			cmd_alert=0;
			cmd_incmd=0;
			cmd_text[0]='\0';
			cmd_final_char=0;
			
			// Set other attribs
			doc_file_open = 1;
			doc_final_char=safestrlen(doc_text);
			
			// Set cursor
			cursor = 0;
			
			break;
	}
}

// Save file
void command_save(chanend from_services) {
	int rtn;
	char filename[MAX_FILENAME_LEN+1];
	
	// Check and see if a filename has been provided
	// Open file
	// Need to validate there's enough space in the filename array
	if (safestrlen(cmd_text)>MAX_FILENAME_LEN+5) {
		command_message(from_services, "Filename too long");
		cmd_alert=1;
		return;
	}
	
	// See if there's anything
	substr(cmd_text, filename, 5, cmd_final_char, 1);
	
	// If there is NO open file, for new file specified
	if ((!doc_file_open) || ((safestrlen(filename)!=0) && (safestrcmp(filename, doc_filename)!=0))) {
		// Validate if we have to use
		if (validate_filename(from_services, filename)!=0) return;
			
		// If exists it's a nogo
		if (services_sdcard_exists(from_services, filename)==0) {
			// Error
			command_message(from_services, "File already exists");
			cmd_alert=1;
			return;
		}
		
		// Copy
		safestrcpy(doc_filename, filename);
		doc_file_open=1;
	}
	
	// Save
	
	// Delete if exists
	services_sdcard_del(from_services, doc_filename);
	
	// Try to write
	rtn = services_sdcard_write(from_services, doc_filename, doc_text, safestrlen(doc_text));

	switch (rtn) {
		case 1:
			command_message(from_services, "SD card not inserted");
			cmd_alert=1;
			break;
		case 2:
			command_message(from_services, "Unable to access file");
			cmd_alert=1;
			break;
		case 3:
			command_message(from_services, "Unable to write entire file ");
			cmd_alert=1;
			break;
		case 4:
			command_message(from_services, "Filename too long");
			cmd_alert=1;
			break;
		default:
			command_message(from_services, "OK!");
			// Rest attribs
			cmd_alert=0;
			cmd_incmd=0;
			cmd_text[0]='\0';
			cmd_final_char=0;
			break;
	}
}

// Load file
// new - clears curr doc and creates new one
// new x - clears curr doc and creates new one called file x
void command_new(chanend from_services) {
	char filename[MAX_FILENAME_LEN+1];
	int rtn;
	
	// Need to validate there's enough space in the filename array
	if (safestrlen(cmd_text)>MAX_FILENAME_LEN+5) {
		command_message(from_services, "Filename too long");
		cmd_alert=1;
		return;
	}
	
	// See if there's anything
	substr(cmd_text, filename, 4, cmd_final_char, 1);
	
	if (validate_filename(from_services, filename)!=0) return;
	
	// Check
	if (filename[0]=='\0') {
		doc_clearstate();
	} else {

		// If exists it's a nogo
		if (services_sdcard_exists(from_services, filename)==0) {
			// Error
			command_message(from_services, "File already exists");
			cmd_alert=1;
		} else {
			// Set things up
			
			// Write a file so no-one else does
			if ((rtn=services_sdcard_write(from_services, filename, "\0", 1))!=0) {
				switch (rtn) {
					case 1:
						command_message(from_services, "SD card not inserted");
						cmd_alert=1;
						break;
					case 2:
						command_message(from_services, "Unable to access file");
						cmd_alert=1;
						break;
					case 3:
						command_message(from_services, "Unable to read print file ");
						cmd_alert=1;
						break;
					case 4:
						command_message(from_services, "Filename too long");
						cmd_alert=1;
						break;
				}
				return;
			}
			
			// Wipe all
			doc_clearstate();
			
			// Set filename
			safestrcpy(doc_filename, filename);
			
			// Set attribs
			doc_file_open=1;
			cmd_incmd=0;
			cmd_text[0]='\0';
			cmd_final_char=0;
		}
	}
}

// Validate a filename, return status and update cmd message
int validate_filename(chanend from_services, char fil[]) {
	int i=0;
	int seen_dot=0;
	
	// Check len
	if (safestrlen(fil)>MAX_FILENAME_LEN) {
		command_message(from_services, "Filename too long");
		cmd_alert=1;
		return 1;
	}
	if (safestrlen(fil)==0) {
		command_message(from_services, "No filename!");
		cmd_alert=1;
		return 1;
	}
	
	// Check all lowercase
	for (i=0; i<safestrlen(fil); i++) {
		if ((fil[i]>0x61) && (fil[i]<0x7B)) {
			command_message(from_services, "Letters must be uppercase");
			cmd_alert=1;
			return 1;
		}
		
		if (fil[i]=='.') seen_dot=i;
	}
	
	// Check for three letter ext
	if (i-seen_dot-1!=3) {
		command_message(from_services, "Please have a three char extension");
		cmd_alert=1;
		return 1;
	}
	
	return 0;
}

// Command message
void command_message(chanend from_services, char text[]) {
	services_graphics_set_line(from_services, cmd_status);
	services_graphics_set_char(from_services, 2);
	services_graphics_add_text(from_services, text);
}

// Clear doc and command state
void doc_clearstate() {
	// Document vars
	safememset(doc_filename, 0, MAX_FILENAME_LEN+1);
	doc_filename[0]='\0';
	safememset(doc_text, 0, MAX_DOC_LEN+1);
	doc_text[0]='\0';
	doc_final_char=0;
	doc_file_open=0;
	cursor=0;		// Cursor is position in the array

	// Commands
	safememset(cmd_text, 0, MAX_EDIT_CMD_LEN + 1);
	cmd_text[0]='\0';
	cmd_final_char=0;
	cmd_alert=0;
	cmd_incmd=0;
	running=1;
}
