// Runs a simple interpreter

// Print the immediate code before execution
//#define SHOW_IMMCODE

// Enable the buttons on the board - warning: this halves the number ops/sec
//#define ENABLE_XGCBUTTONS

// In this case the side-effect of having things in global vars is useful - we can persist
// the state accross different different calls, to create an interactive interpreter.

#include "xgc_interpreter.h"

// Settings
#define MAX_VAR_NAME_LEN 10			// Max variable name length
#define MAX_TOKEN_LEN 50			// Max length of a command token
#define NO_OF_VARS 30				// Number of variables allocated initially
#define NO_OF_LOOPS 20				// Number of nested loops and conditionals
#define MAX_TOKENS 250				// Maximum number of tokens in an entire program

// System registers
// There will be 4 system registers available in sysstore which will be allocated by
// hand for use in instructions that span multiple tokens and require temporary vars.
// The system result register shall contain the result of all arithmetic and boolean
// operations and should be defined after these are executed.
// **At all other times the value of the system registers will be undefined.**
// The input device register shall contain if a button or input device was pressed.
//The machine will automatically update this register every time an instruction is executed.
// In order:
// 0|JR|JL|JD|JU|JB|BC|BB|BA
#define SYSVAR_RESULT 2				// Which system register is the result register
#define SYSVAR_INPUT 3				// Which system register is the input register

// Tokens
// There are two types of tokens. Text tokens are the commands the user can give to the
// interpreter and machine tokens are tokens in intermediate code which are executed.
// Text toks
typedef enum text_tok {
	TTOK_BEEP,					// Text token BEEP			beep [1/10sec]@[freq Hz]
	TTOK_SLEEP,					// Text token SLEEP			sleep [1/10sec]
	TTOK_SET,					// Text token SET			set x=[arithexp] where arithexp=(xopy)
	TTOK_PRINT,					// Text token PRINT			print var
	TTOK_IF,					// Text token IF			if boolexp {block} where boolexp=(xopy)
	TTOK_ELSE,					// Text token ELSE			else when encased in IF and ENDIF. Indicates ELSE portion of IF.
	TTOK_ENDIF,					// Text token ENDIF			endif when IF has been used previously. Indicates end of IF block.
	TTOK_WHILE,					// Text token WHILE			while boolexp {block} when boolexp=(xopy)
	TTOK_ENDWHILE,				// Text token ENDWHILE		endwhile when WHILE has been used previously. Indicates end of WHILE block.
	TTOK_PRINTLN,				// Text token PRINTLN		println prints a new line
	TTOK_SETCUR,				// Text token SETCUR		setcur x@y sets the graphics cursor to the specified X and Y co-ordinates
	TTOK_GETKEY,				// Text token GETKEY		getkey var	Gets a key from the keyboard buffer and stores in var. If no key is available, returns -1.
	TTOK_RANDOM,				// Text token RANDOM		random dest max - Generates a "random" number between 0 and max
	TTOK_PRINTNC				// Text token PRINTNC		printnc int - Interpret an integer as a char and print it
} text_tok;
// Machine toks
typedef enum machine_tok {
	MTOK_BEEP,					// Machine token BEEP		v1=[1/10sec] v2=Hz op=Undefined						Beeps for v1 at a freq of v2Hz
	MTOK_SLEEP,					// Machine token SLEEP		v1=[1/10sec] v2=Undefined op=Undefined				Sleeps for v1
	MTOK_SET,					// Machine token SET		v1=to-set-var(IS_VARIABLE) v2=value	op=Undefined	Performs to-set-var=value
	MTOK_PRINT,					// Machine token PRINT		v1=var v2=Undefined op=Undefined					Prints var, char or -1 for new line to screen
	MTOK_ARITH,					// Machine token ARITH		v1=x v2=y op=arithmetic-op							Performs arithmetic expression SYSRES=X-op-Y
	MTOK_BOOLEXP,				// Machine token BOOLEXP	v1=x v2=y op=boolean-op								Performs a boolean expression SYSRES=X-op-Y
	MTOK_JMPIFF,				// Machine token JMPIFF		v1=x v2=token-postition op=Undefined				Performs a conditional jump: if x is false then goto token-position.
	MTOK_SETCUR,				// Machine token SETCUR		v1=x-coord v2=y-co-ord op=Undefined					Sets the cursor position to x,y
	MTOK_GETKEY,				// Machine token GETKEY		v1=var-to-store-in v2=Undefined op=Undefined
	MTOK_RANDOM,				// Machine token RANDOM		v1=var-to-store-in v2=max-int op=Undefined
	MTOK_PRINTNC				// Machine token PRINTNC	v1=var-or-int-to-print v2=Undefined op=Undefined	Interprets a variable or int as a character and prints it.
} machine_tok;


// A way of indicating if an int is a variable, integer or result register
typedef enum { IS_INTEGER, IS_VARIABLE, IS_SYSVAR } use_type;
// Operators
typedef enum { ASSIGN, EQUAL, NOTEQUAL, LESSTHAN, GREATERTHAN, LESSEQUAL, GREATEREQUAL, PLUS, MINUS, MULTIPLY, DIVIDE, AT, AND, OR, BITWISE_AND, BITWISE_OR } op_type;

// Struct for storing vars
typedef struct variable {
	char name[MAX_VAR_NAME_LEN+1];
	int value;
} variable;

// Struct for storing tokens
typedef struct token {
	machine_tok type;
	int value;
	use_type val_type;
	int operation;
	int value2;
	use_type val2_type;
} token;

// Utils
void remove_whitespace(char command[]);
extern int string2int(char str[]);
int is_operator(char i);
int string2varid(char name[]);
int add_token(chanend from_services, machine_tok tok_name, int v1, use_type u1, int v2, use_type u2, int op, int token_ptr);
int get_value(int pos, int val);
void set_value(int pos, int val, int thing);
// Main
void interpret(chanend from_services, char command[]);
void preprocess(chanend from_services, char command[]);
void execute(chanend from_services);
void update_input_status_reg(chanend from_services);
// Tokens
{ text_tok, int } extract_token(chanend from_services, int token_ptr, char extracted_token[]);
int text_to_machine(chanend from_services, int token_ptr);
{ int, use_type, int } expression_gen(chanend from_services, int token_ptr, char param[]);
{ int, use_type, op_type, int, use_type, int } split_op_argument(chanend from_services, char param[], int token_ptr);
int bool_expression_gen(char exp[], int token_ptr, chanend from_services);
int special_id_to_int(chanend from_services, char name[]);
// Actual functions
int tok_beep(int pos, chanend from_services);
int tok_sleep(int pos, chanend from_services);
int tok_do(int pos, chanend from_services);
int tok_repeat(int pos, chanend from_services);
int tok_set(int pos, chanend from_services);
int tok_print(int pos, chanend from_services);
int tok_arith(int pos, chanend from_services);
int tok_boolexp(int pos, chanend from_services);
int tok_jmpiff(int pos, chanend from_services);
int tok_setcur(int pos, chanend from_services);
int tok_getkey(int pos, chanend from_services);
int tok_random(int pos, chanend from_services);
int tok_printnc(int pos, chanend from_services);

// Used in the execution stage
//-------------------------------

// Provide a store of variables
variable store[NO_OF_VARS];
static int max_vars=0;

// Provide a store of system vars for use in evaluating expressions
// Note that each statement can only be broken into 2 sep expressions
// 0 - Exp1, 1 - Exp2, 2 - Result, 3 - Input status
variable sysstore[4];

// Used in both stages
// --------------------------

// Provide a store of tokens
token toks[MAX_TOKENS];
static int max_toks=0;

// #defines for true and false
#define BOOL_TRUE 1
#define BOOL_FALSE 0

// Used in preprocessing only
// ---------------------------------

// Was there an error in preprocessing?
int error=0;

// The current token type (TTOK) and the extracted parameter
// In the future would like just to use a global array to reduce memory.
text_tok curr_token;
char extracted_param[MAX_TOKEN_LEN+1];

// This is a stack for preprocessing of loops and ifs
// I decided to move all loop resolution into the preprocessor in order
// to make the intermediate code more simple
int ppstack[NO_OF_LOOPS];
static int pps_next=0;

// SPECIAL IDS
// Left hand side special ids
	// We can refer to bits in the SYSREG_INPUT register by ANDing using these #defines
	// Can also refer to buttons if required by converting their text names into the integers below.
	// The button names are known as "special ids".
	// The special ids are preprocessed out into an AND instruction along with the value.
	#define BUTTON_A (0x00000001)
	#define BUTTON_B (0x00000002)
	#define BUTTON_C (0x00000004)
	#define BUTTON_JC (0x00000008)
	#define JOY_UP (0x00000010)
	#define JOY_DOWN (0x00000020)
	#define JOY_LEFT (0x00000040)
	#define JOY_RIGHT (0x00000080)
// Right hand side special ids
	// We can also provide some special ids which change into integers
	// which are not ANDed. These are:
	// KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT, KEY_ENTER, KEY_SPACE

// Overall interpreter functions
// ---------------------------------

// Function for embedding into shell
void interpreter_embed(chanend from_services, char command[]) {
	error=0;
	
	// Stages
	// Strip whitespace, preprocess into tokens
	preprocess(from_services, command);

	// Set the keyboard to non-blocking mode
	from_services <: AK_KEY_SETMODE;
	from_services <: 1;
	
	// Execute tokens
	if (error!=1) execute(from_services);
	
	// Set the keyboard to blocking mode
	from_services <: AK_KEY_SETMODE;
	from_services <: 0;
}

// Function for running from SDC taking over the thread
void interpreter_full(chanend from_services, char command[]) {
	
	// If we're running a large SDC program, reset the state
	reset_state();
	
	// Display a waiting message
	services_graphics_add_text(from_services, "Loading: Please wait.");
	services_graphics_new_line(from_services);
	services_graphics_new_line(from_services);
	
	// Stages
	// Strip whitespace, preprocess into tokens
	preprocess(from_services, command);

	// Pause if errors
	if (error) {
		services_graphics_new_line(from_services);
		services_graphics_add_text(from_services, "Error(s) found. Press a key to exit.");
		from_services <: AK_KEY_REQ;
		from_services :> int dump;
	} else {
		// Set the keyboard to non-blocking mode
		from_services <: AK_KEY_SETMODE;
		from_services <: 1;
		
		// Clear screen
		services_graphics_cs(from_services);
		
		// Execute tokens
		execute(from_services);
		
		// Set the keyboard to blocking mode
		from_services <: AK_KEY_SETMODE;
		from_services <: 0;
	}
}

// Reset all the state
void reset_state() {
	max_vars=0;
	max_toks=0;
	error=0;
	pps_next=0;
}

// Execution functions
// -----------------------------------------------------------------------------

// Execute the stored tokens
void execute(chanend from_services) {
	int i;
#ifdef SHOW_IMMCODE
	// Print out the program in terms of tokens
	for (i=0; i<max_toks; i++) {
		// Print line
		printint(i);
		printchar(' ');
		
		// Print token name
		switch (toks[i].type) {
			case MTOK_BEEP:
				printstr("beep");
				break;
			case MTOK_SLEEP:
				printstr("sleep");
				break;
			case MTOK_SET:
				printstr("set");
				break;
			case MTOK_PRINT:
				printstr("print");
				break;
			case MTOK_ARITH:
				printstr("arith");
				break;
			case MTOK_BOOLEXP:
				printstr("boolexp");
				break;
			case MTOK_JMPIFF:
				printstr("jmpiff");
				break;
			case MTOK_SETCUR:
				printstr("setcur");
				break;
			case MTOK_GETKEY:
				printstr("getkey");
				break;
			case MTOK_RANDOM:
				printstr("random");
				break;
			case MTOK_PRINTNC:
				printstr("printnc");
				break;
			default:
				printstr("UNKNOWN");
				break;
		}
		printchar(' ');
		
		// Print params
		printint(toks[i].value);
		switch (toks[i].val_type) {
			case IS_INTEGER:
				printchar('i');
				break;
			case IS_SYSVAR:
				printchar('s');
				break;
			case IS_VARIABLE:
				printchar('v');
				break;
			default:
				printchar('!');
				break;
		}
		
		// Print params
		printchar(' ');
		printint(toks[i].value2);
		switch (toks[i].val2_type) {
			case IS_INTEGER:
				printchar('i');
				break;
			case IS_SYSVAR:
				printchar('s');
				break;
			case IS_VARIABLE:
				printchar('v');
				break;
			default:
				printchar('!');
				break;
		}
		
		// And then the op
		printchar(' ');
		if (toks[i].operation!=0) printint(toks[i].operation);
		
		printstrln("");
	}
#endif
	
#ifdef ENABLE_XGCBUTTONS
	// Perform an intial update of the input status register
	update_input_status_reg(from_services);
#endif
	
	for (i=0; i<max_toks;) {
		// Process each tok in turn
		switch (toks[i].type) {
			case MTOK_BEEP:
				i = tok_beep(i, from_services);
				break;
			case MTOK_SLEEP:
				i = tok_sleep(i, from_services);
				break;
			case MTOK_SET:
				i = tok_set(i, from_services);
				break;
			case MTOK_PRINT:
				i = tok_print(i, from_services);
				break;
			case MTOK_ARITH:
				i = tok_arith(i, from_services);
				break;
			case MTOK_BOOLEXP:
				i = tok_boolexp(i, from_services);
				break;
			case MTOK_JMPIFF:
				i = tok_jmpiff(i, from_services);
				break;
			case MTOK_SETCUR:
				i = tok_setcur(i, from_services);
				break;
			case MTOK_GETKEY:
				i = tok_getkey(i, from_services);
				break;
			case MTOK_RANDOM:
				i = tok_random(i, from_services);
				break;
			case MTOK_PRINTNC:
				i = tok_printnc(i, from_services);
				break;
		}
		
#ifdef ENABLE_XGCBUTTONS
		// Update the status of the joystick and buttons
		update_input_status_reg(from_services);
#endif
	}
}

// Interprets a variable or int as a character
int tok_printnc(int pos, chanend from_services) {
	int val = get_value(pos, 1);

	if ((val>=0)&&(val<=0x7E)) services_graphics_add_char(from_services, val);
	
	return ++pos;
}

// Generate a "random" number
int tok_random(int pos, chanend from_services) {
	unsigned int y;
	int t;
	timer tme;
	
	// Get time
	tme :> t;
	
	// Mangle it
	crc32(y, t, 0xBA0DC66B);
	t=y;
	crc32(y, t, 0xEB31D82E);
	
	// Mod
	y = y%(get_value(pos, 2)+1);
	
	// And set
	set_value(pos, 1, y);
	
	return ++pos;
}

// Provide a non-blocking keyboard routine, set var with char if
// available in the buffer, otherwise return -1
int tok_getkey(int pos, chanend from_services) {
	int key;
	
	from_services <: AK_KEY_NBREQ;
	from_services :> key;
	
	set_value(pos, 1, key);
	
	return ++pos;
}

// Update the status of the joystick and buttons
void update_input_status_reg(chanend from_services) {
	from_services <: AK_GET_INPUT_STATUS;
	from_services :> sysstore[3].value;
}

// Sets the position of the cursor
int tok_setcur(int pos, chanend from_services) {
	int x = get_value(pos, 1);
	int y = get_value(pos, 2);
	
	from_services <: AK_GRAPHICS_SET_CHAR;
	from_services <: x;
	
	from_services <: AK_GRAPHICS_SET_LINE;
	from_services <: y;
	
	return ++pos;
}

// Print the contents of a var
int tok_print(int pos, chanend from_services) {
	int val;
	
	// Print
	if(toks[pos].val_type==IS_VARIABLE) {
		// Convert to chars
		char tmp[MAX_TOKEN_LEN+1];

		// Convert to string
		itoa(get_value(pos, 1), tmp);
				
		services_graphics_add_text(from_services, tmp);
	} else if (toks[pos].val_type==IS_INTEGER) {
		val = get_value(pos, 1);
		
		if (val==-1) {
			// New line
			services_graphics_new_line(from_services);
		} else {
			// Interpret int as char
			services_graphics_add_char(from_services, val);
		}
	}
	
	return ++pos;
}

// Set, simply write the variable
int tok_set(int pos, chanend from_services) {
	int destvar = toks[pos].value;
	
	// Do assignment
	if (toks[pos].val_type==IS_SYSVAR) {
		sysstore[destvar].value = get_value(pos, 2);
	} else {
		store[destvar].value = get_value(pos, 2);
	}

	return ++pos;
}

// Stop
int tok_sleep(int pos, chanend from_services) {
	// So do a sleep for so many 1/10 secs
	int length = toks[pos].value*10000000;
	timer t;
	int tme;

	t :> tme;
	
	tme += length;

	t when timerafter(tme) :> void;
	
	return ++pos;
}

// Do a beep
int tok_beep(int pos, chanend from_services) {

	// Check the value of pos2 to make sure it's not zero
	if (get_value(pos, 2)==0) {
		services_graphics_add_text(from_services, "Run time error: Parameter in BEEP cannot be zero at runtime.");
		services_graphics_new_line(from_services);
		
		return ++pos;
	}
	
	// Send to services
	from_services <: AK_SPEAKER_TONE;
	from_services <: get_value(pos, 1);
	from_services <: get_value(pos, 2);
	
	return ++pos;
}

// Arithmetic operation, result is stored in sysstore[2]
int tok_arith(int pos, chanend from_services) {
	int first, second, result;
	
	// Load value for calc
	first = get_value(pos, 1);
	second = get_value(pos, 2);
	
	// Do operation
	switch (toks[pos].operation) {
		case PLUS:
			result = first + second;
			break;
		case MINUS:
			result = first - second;
			break;
		case MULTIPLY:
			result = first * second;
			break;
		case DIVIDE:
			// Check the value of pos2 to make sure it's not zero
			if (second==0) {
				services_graphics_add_text(from_services, "Run time error: Division by zero detected.");
				services_graphics_new_line(from_services);
				
				return pos;
			}
			
			result = first / second;
			break;
		case BITWISE_AND:
			result = (first & second);
			break;
		case BITWISE_OR:
			result = (first | second);
			break;
	}
	
	// Store in SYSVAR2
	sysstore[SYSVAR_RESULT].value = result;
	
	// Return
	return ++pos;
}

// Boolean operation, result is stored in sysstore[2]
int tok_boolexp(int pos, chanend from_services) {
	int first, second, result;
	
	// Load value for calc
	first = get_value(pos, 1);
	second = get_value(pos, 2);
	
	// Do operation
	switch (toks[pos].operation) {
		case EQUAL:
			result = (first==second);
			break;
		case NOTEQUAL:
			result = (first!=second);
			break;
		case LESSTHAN:
			result = (first<second);
			break;
		case GREATERTHAN:
			result = (first>second);
			break;
		case LESSEQUAL:
			result = (first<=second);
			break;
		case GREATEREQUAL:
			result = (first>=second);
			break;
		case AND:
			result = (first && second);
			break;
		case OR:
			result = (first || second);
			break;
	}
	
	// Store in SYSVAR2
	sysstore[SYSVAR_RESULT].value = result;
	
	// Return
	return ++pos;
}

// Jumps to a location if the value indicated is ZERO [BOOL_FALSE]
// v1 should be the test register and v2 should be where to jump
int tok_jmpiff(int pos, chanend from_services) {
	int first, second;
	
	// Load values
	first = get_value(pos, 1);
	second = get_value(pos, 2);
	
	// Jump
	if (first==BOOL_FALSE) return second; else return ++pos;
}

// Set value
void set_value(int pos, int val, int thing) {
	if (val==1) {
		use_type type = toks[pos].val_type;
			
		switch (type) {
			case IS_VARIABLE:
				store[toks[pos].value].value = thing;
				break;
			case IS_SYSVAR:
				sysstore[toks[pos].value].value = thing;
				break;
		}
	} else {
		use_type type = toks[pos].val2_type;
			
		switch (type) {
			case IS_VARIABLE:
				store[toks[pos].value2].value = thing;
				break;
			case IS_SYSVAR:
				sysstore[toks[pos].value2].value = thing;
				break;
		}
	}
}

// Take a value and type and return abs value
int get_value(int pos, int val) {
	if (val==1) {
		use_type type = toks[pos].val_type;
			
		switch (type) {
			case IS_INTEGER:
				return toks[pos].value;
			case IS_VARIABLE:
				return store[toks[pos].value].value;
			case IS_SYSVAR:
				return sysstore[toks[pos].value].value;
		}
	} else {
		use_type type = toks[pos].val2_type;
			
		switch (type) {
			case IS_INTEGER:
				return toks[pos].value2;
			case IS_VARIABLE:
				return store[toks[pos].value2].value;
			case IS_SYSVAR:
				return sysstore[toks[pos].value2].value;
		}
	}
	
	return 0;
}
// Preprocess functions
// -----------------------------------------------------------------------------

// Preprocess
void preprocess(chanend from_services, char command[]) {
	// Counter
	int i=0;
	// Previous token
	int extoken_prv=0;
	// Pointer to token array
	int token_ptr=0;
	// Current token and has parameters?
	text_tok curr_tok_type;
	int curr_tok_param=0;
	
	// Check for numbers of ifs/endifs/whiles/endwhiles
	int no_ifs=0;
	int no_whiles=0;
	
	// Remove excess whitespace
	remove_whitespace(command);

	// Start at the front, work along until find space, extract and then see if it matches a token
	for (i=0; i<=safestrlen(command); i++) {
		
		// If the current char is a space, split
		if ((command[i]==' ') || (i==safestrlen(command))) {
			// Check length
			if ((i-extoken_prv)>MAX_TOKEN_LEN) {
				services_graphics_add_text(from_services, "Error: Single command is longer than max.");
				services_graphics_new_line(from_services);
				error=1;
				break;
			}
			
			// If this is the first token (the command) then store it
			if (curr_tok_param==0) {
				// Extract command
				substr(command, extracted_param, extoken_prv, i, 1);
				
				// Store the command and if there are any params
				{ curr_token, curr_tok_param }  = extract_token(from_services, token_ptr, extracted_param);

				// Check for invalid
				if (curr_tok_type==-1) {
					// Reset
					curr_token=0;
					curr_tok_param=0;
				} else {
					// Set parameter array to zero
					extracted_param[0] = '\0';
					
					// If there are no params then process into machine token
					if (!curr_tok_param) {
						// Process
						token_ptr = text_to_machine(from_services, token_ptr);
					}
				}

				// Set end
				extoken_prv = i+1;
			} else if (curr_tok_param==1) {
				// If we have params
				
				// Extract params
				substr(command, extracted_param, extoken_prv, i, 1);
				
				// Store them
				token_ptr = text_to_machine(from_services, token_ptr);
						
				// Set end
				extoken_prv = i+1;
				
				// Reset token no
				curr_token=0;
				curr_tok_param=0;
			}
			
			// Check numbers of whiles/ifs
			switch (curr_token) {
				case TTOK_WHILE:
					no_whiles++;
					break;
				case TTOK_ENDWHILE:
					no_whiles--;
					break;
				case TTOK_IF:
					no_ifs++;
					break;
				case TTOK_ENDIF:
					no_ifs--;
					break;
				default:
					break;
			}
		}
		
		// Check range of prog
		if (token_ptr>=MAX_TOKENS) {
			services_graphics_add_text(from_services, "Error: Program is too big.");
			services_graphics_new_line(from_services);
			error=1;
			return;
		}
	}

	// Check whiles/ifs
	if (no_whiles!=0) {
		services_graphics_add_text(from_services, "Warning: WHILE mismatch - probably missing ENDWHILE.");
		services_graphics_new_line(from_services);
		error=1;
	}
	if (no_ifs!=0) {
		services_graphics_add_text(from_services, "Warning: IF mismatch - probably missing ENDIF.");
		services_graphics_new_line(from_services);
		error=1;
	}
	
	// Set max
	max_toks = token_ptr;
}

// Store a token stem in the token array 
// IMPORTANT: if token expects parameters, should return 1
// if token does not expect parameters should return 0
{ text_tok, int } extract_token(chanend from_services, int token_ptr, char extracted_token[]) {
	if (safestrcmp(extracted_token, "beep")==0) {
		return { TTOK_BEEP, 1 };
	} else if (safestrcmp(extracted_token, "sleep")==0) {
		return { TTOK_SLEEP, 1 };
	} else if (safestrcmp(extracted_token, "set")==0) {
		return { TTOK_SET, 1 };
	} else if (safestrcmp(extracted_token, "print")==0) {
		return { TTOK_PRINT, 1 };
	} else if (safestrcmp(extracted_token, "if")==0) {
		return { TTOK_IF, 1 };
	} else if (safestrcmp(extracted_token, "else")==0) {
		return { TTOK_ELSE, 0 };
	} else if (safestrcmp(extracted_token, "endif")==0) {
		return { TTOK_ENDIF, 0 };
	} else if (safestrcmp(extracted_token, "while")==0) {
		return { TTOK_WHILE, 1 };
	} else if (safestrcmp(extracted_token, "endwhile")==0) {
		return { TTOK_ENDWHILE, 0 };
	} else if (safestrcmp(extracted_token, "println")==0) {
		return { TTOK_PRINTLN, 0 };
	} else if (safestrcmp(extracted_token, "setcur")==0) {
		return { TTOK_SETCUR, 1 };
	} else if (safestrcmp(extracted_token, "getkey")==0) {
		return { TTOK_GETKEY, 1 };
	} else if (safestrcmp(extracted_token, "random")==0) {
		return { TTOK_RANDOM, 1 };
	} else if (safestrcmp(extracted_token, "printnc")==0) {
		return { TTOK_PRINTNC, 1 };
	} else {
		services_graphics_add_text(from_services, "Warning: Invalid token ");
		services_graphics_add_text(from_services, extracted_token);
		services_graphics_new_line(from_services);
		error=1;
	}

	// No token, no params
	return { -1, -1 };
}

// Process a token from the source file into a set of executable tokens. There may
// not be a 1-to-1 mapping between a source token and the destination executable tokens.
// Should return final token_ptr ++
int text_to_machine(chanend from_services, int token_ptr) {
	char tmp[MAX_TOKEN_LEN+1];
	int tmpp=0;

	// Range check for stack pointer used in IFs and WHILEs
	if ((pps_next==0) && ((curr_token==TTOK_ENDIF) || (curr_token==TTOK_ENDWHILE) || (curr_token==TTOK_ELSE))) {
		services_graphics_add_text(from_services, "Error: Fatal syntax error, suspect missing ENDIF or ENDWHILE.");
		services_graphics_new_line(from_services);
		error=1;
		
		return token_ptr;
	}
	
	// Process token params
	if (curr_token==TTOK_SLEEP) {
		int v1;
		use_type u1;
		
		// Use expressiongen
		{ v1, u1, token_ptr }  = expression_gen(from_services, token_ptr, extracted_param);

		// Store param as int
		toks[token_ptr].type = MTOK_SLEEP;
		toks[token_ptr].value = v1;
		toks[token_ptr].val_type = u1;

		// Check
		if (toks[token_ptr].value<0) {
			services_graphics_add_text(from_services, "Warning: Negative integer detected at ");
			services_graphics_add_text(from_services, extracted_param);
			services_graphics_new_line(from_services);
			error=1;
		}
		
		// Return tok_ptr++
		return ++token_ptr;
	} else if ((curr_token==TTOK_BEEP) || (curr_token==TTOK_SETCUR)) {
		int tv;
		use_type tu;
		
		int at_sign;
		
		// Find the at sign
		at_sign = safestrchr(extracted_param, '@');
		// If not AT then prob
		if (at_sign==-1) {
			if (curr_token==TTOK_BEEP) services_graphics_add_text(from_services, "Error: BEEP expects [1/10 sec]@[y Hz]. If in doubt use @1000 .");
			if (curr_token==TTOK_SETCUR) services_graphics_add_text(from_services, "Error: SETCUR expects x@y.");
			
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}
		
		// Extract the param
		substr(extracted_param, tmp, 0, at_sign, 1);
		{ tv, tu, token_ptr } = expression_gen(from_services, token_ptr, tmp);
		// Move to SYSVAR0
		token_ptr = add_token(from_services, MTOK_SET, 0, IS_SYSVAR, tv, tu, ASSIGN, token_ptr);
		// Check for error
		if ((tu==IS_INTEGER) && (tv<=0) && (curr_token==TTOK_BEEP)) {
			services_graphics_add_text(from_services, "Error: Neither BEEP argument can be zero.");
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}
		// Extract next and repeat and store in SYSVAR1
		substr(extracted_param, tmp, at_sign+1, safestrlen(extracted_param)+1, 1);
		
		{ tv, tu, token_ptr } = expression_gen(from_services, token_ptr, tmp);
		token_ptr = add_token(from_services, MTOK_SET, 1, IS_SYSVAR, tv, tu, ASSIGN, token_ptr);
		// Check for error
		if ((tu==IS_INTEGER) && (tv<=0) && (curr_token==TTOK_BEEP)) {
			services_graphics_add_text(from_services, "Error: Neither BEEP argument can be zero.");
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}
		
		// Add the token
		if (curr_token==TTOK_BEEP) {
			return add_token(from_services, MTOK_BEEP, 0, IS_SYSVAR, 1, IS_SYSVAR, 0, token_ptr);
		} else {
			return add_token(from_services, MTOK_SETCUR, 0, IS_SYSVAR, 1, IS_SYSVAR, 0, token_ptr);
		}
	} else if (curr_token==TTOK_PRINT) {
		// Print will accept a char in '' or a variable name
		
		// Check
		if ((!((extracted_param[0]>=0x61) && (extracted_param[0]<=0x7A))) && (extracted_param[0]!='\'')) {
			services_graphics_add_text(from_services, "Warning: PRINT expects a variable or a char encased in ''.");
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}
		
		// Do char
		if (extracted_param[0]=='\'') {
			// Set var and type
			toks[token_ptr].type = MTOK_PRINT;
			toks[token_ptr].value = extracted_param[1];
			toks[token_ptr].val_type = IS_INTEGER;
		} else {
			// Set var and type
			toks[token_ptr].type = MTOK_PRINT;
			toks[token_ptr].value = string2varid(extracted_param);
			toks[token_ptr].val_type = IS_VARIABLE;
		}
		
		// And return
		return ++token_ptr;
	} else if (curr_token==TTOK_PRINTLN) {
		// Print a new line
		toks[token_ptr].type = MTOK_PRINT;
		toks[token_ptr].value = -1;
		toks[token_ptr].val_type = IS_INTEGER;
		
		return ++token_ptr;
	} else if (curr_token==TTOK_SET) {
		// A set is the phrase [varname]=[exp] where an exp is:
		// - An integer
		// - A variable
		// - A bracketed arithmetic expression of the format ([var|int] op [var|int])
		int i;
		
		int v1, v2;
		use_type u1, u2;

		// Check is varname
		if (!((extracted_param[0]>=0x61) && (extracted_param[0]<=0x7A))) {
			// Error
			services_graphics_add_text(from_services, "Warning: Invalid assignment on LHS.");
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}

		// Copy
		for (i=0;i<MAX_VAR_NAME_LEN;i++) {
			if (extracted_param[i]!='=') {
				tmp[tmpp] = extracted_param[i];
				tmpp++;
			} else {
				// Add null
				tmp[tmpp] = '\0';
				
				// Exit
				break;
			}
		}

		// See if it reached the end
		if (i==MAX_VAR_NAME_LEN) {
			services_graphics_add_text(from_services, "Error: Expected SET is x=y.");
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}
		
		// Set the first arg
		v1 = string2varid(tmp);
		u1 = IS_VARIABLE;
		
		// Use the rest to pass to the expression parser, which will return something that is appropriate
		// for use in v2, u2.
		substr(extracted_param, tmp, i+1, safestrlen(extracted_param)+1, 1);

		{ v2, u2, token_ptr }  = expression_gen(from_services, token_ptr, tmp);

		// Now assign all
		token_ptr = add_token(from_services, MTOK_SET, v1, u1, v2, u2, ASSIGN, token_ptr);

		// Sanity checks
		if (toks[token_ptr-1].val_type==IS_INTEGER) {
			services_graphics_add_text(from_services, "Warning: Invalid assignment on LHS.");
			services_graphics_new_line(from_services);
			error=1;
		}
		if ((toks[token_ptr-1].val_type==toks[token_ptr-1].val2_type) && (toks[token_ptr-1].value==toks[token_ptr-1].value2)) {
			services_graphics_add_text(from_services, "Warning: Pointless assignment detected.");
			services_graphics_new_line(from_services);
			error=1;
		}
		
		return token_ptr;
	} else if ((curr_token==TTOK_IF) || (curr_token==TTOK_WHILE)) {
		// Check boundaries
		int oob=0;
		if (curr_token==TTOK_IF) {
			if (pps_next+1>NO_OF_LOOPS) {
				oob=1;
			}
		}
		if (curr_token==TTOK_WHILE) {
			if (pps_next+2>NO_OF_LOOPS) {
				oob=1;
			}	
		}
		if (oob) {
			services_graphics_add_text(from_services, "Error: Run out of conditional variable stackspace.");
			services_graphics_new_line(from_services);
			error=1;
			return ++token_ptr;
		}
		
		// Put position to jump back to on
		if (curr_token==TTOK_WHILE) { ppstack[pps_next++] = token_ptr; }
		
		// In an IF, evaluate expression and then push this position onto the stack
		token_ptr = bool_expression_gen(extracted_param, token_ptr, from_services);
		
		// Add a token
		// v1 is value to look at, v2 is where to jump
		token_ptr = add_token(from_services, MTOK_JMPIFF, SYSVAR_RESULT, IS_SYSVAR, -1, IS_INTEGER, 0, token_ptr);
		
		// Put position on
		ppstack[pps_next++] = token_ptr-1;
	} else if (curr_token==TTOK_ELSE) {
		// We also need to insert a jump to the end of the IF statement
		token_ptr = add_token(from_services, MTOK_JMPIFF, BOOL_FALSE, IS_INTEGER, -1, IS_INTEGER, 0, token_ptr);
		
		// Update the IF statement
		toks[ppstack[--pps_next]].value2 = token_ptr;
		
		// And push it onto the stack
		ppstack[pps_next++] = token_ptr-1;
	} else if (curr_token==TTOK_ENDIF) {
		// Update either the IF token or the "just before else" token
		toks[ppstack[--pps_next]].value2 = token_ptr;

	} else if (curr_token==TTOK_ENDWHILE) {
		int while_cond_pos;
		int while_jmp_pos;
		
		// Dec
		while_cond_pos = ppstack[--pps_next];
		while_jmp_pos = ppstack[--pps_next];
		
		// Add a statement to jump to the start
		token_ptr = add_token(from_services, MTOK_JMPIFF, BOOL_FALSE, IS_INTEGER, while_jmp_pos, IS_INTEGER, 0, token_ptr);
		
		// Update that it should jump here on false
		toks[while_cond_pos].value2 = token_ptr;
	} else if (curr_token==TTOK_GETKEY) {
		// Actually we can't really do any preprocessing on this command (it's more useful to use the
		// underlying services)
		token_ptr = add_token(from_services, MTOK_GETKEY, string2varid(extracted_param), IS_VARIABLE, -1, -1, 0, token_ptr);
	} else if (curr_token==TTOK_RANDOM) {
		// Again, this translates directly to an underlying command
		// however we must extract the var name and the random "between" can be an expression
		int comma;
		
		// Uses for first
		use_type use1, use2;
		int val1, val2;
		
		// Find the at sign
		comma = safestrchr(extracted_param, ',');
		// If not AT then prob
		if (comma==-1) {
			services_graphics_add_text(from_services, "Error: RANDOM expectes dest,max_no.");
			services_graphics_new_line(from_services);
			error=1;
			
			return token_ptr;
		}
		
		// Extract the param
		substr(extracted_param, tmp, 0, comma, 1);

		// Save
		use1 = IS_VARIABLE;
		val1 = string2varid(tmp);

		// Extract next and repeat and store in SYSVAR1
		substr(extracted_param, tmp, comma+1, safestrlen(extracted_param)+1, 1);
		
		{ val2, use2, token_ptr } = expression_gen(from_services, token_ptr, tmp);
		
		token_ptr = add_token(from_services, MTOK_RANDOM, val1, use1, val2, use2, 0, token_ptr);
	} else if (curr_token==TTOK_PRINTNC) {
		// Printnc can evaluate expressions
		
		// Get expression result
		use_type u1; int v1;
		{ v1, u1, token_ptr } = expression_gen(from_services, token_ptr, extracted_param);

		// And store
		token_ptr = add_token(from_services, MTOK_PRINTNC, v1, u1, -1, -1, 0, token_ptr);
	}
	
	// Else skip
	return token_ptr;

}

// Generate code to evaluate a boolean expression, storing the result in the result
// register SYSRES. The formula is to evaluate an expression, move the result from
// SYSRES to SYS0 and perform the next expression. Once we have completed two
// expressions evaulate them together and then evaluate subsequent expressions with the result
// of the previous ones - i.e SYSRES=next_res, SYSRES=SYSRES op SYS0, SYS0=SYSRES.
// When we have finished evaluating expressions then the result will be in SYS0.
// (We will then move it to SYSRES).

// The format of expressions will be strict ((exp1)op(exp2)op(exp3)...(expn)) and no nesting
// will be allowed, where op can be AND or OR (presented here as & and |).
// If an expression fails to be parsed correctly, an error will be generated.

// Externally, the language will have no comprehension of true and false, so a user
// will have to use if (x==1).

// User can assume that SYSRES contains the result of the boolean expression.

int bool_expression_gen(char exp[], int token_ptr, chanend from_services) {
    char tmp[MAX_TOKEN_LEN+1];
    int tmpp=0;
    
    int no_open=0;
    int no_close=0;
    
    int i;
    
    // Do some preliminary checking on the brackets
    for(i=0; i<safestrlen(exp); i++) {
       if (exp[i]=='(') {
            no_open++;
       } else if (exp[i]==')') {
            no_close++;
       }
    }

    if ((no_open!=no_close) || (exp[safestrlen(exp)-1]!=')') || (exp[0]!='(')) {
        services_graphics_add_text(from_services, "Error: Brace mismatch in boolean expression.");
		services_graphics_new_line(from_services);
		error=1;
		
        return token_ptr;
    }
    
    // Since we need to pass through the subexpressions, inc brackets, to expression_gen we
    // should just strip out the first/last set of brackets and extract the subexpressions
    
    // First one is special case - see if we only have one expression
    if (!((exp[0]=='(') && (exp[1]=='('))) {
    	// Only single expression
    	
    	// Easy to evaluate!!!!!!
    	{ void, void, token_ptr }  = expression_gen(from_services, token_ptr, exp);
    	
    	return token_ptr;
    } else {
    	int operator=0;
    	
    	// Several expressions

    	// We always need to do a pair of expressions for AND and OR
    	// Set the register to TRUE
    	token_ptr = add_token(from_services, MTOK_SET, 0, IS_SYSVAR, 1, IS_INTEGER, ASSIGN, token_ptr);
    	
    	// Go through string and do evaluation on close bracket, for the moment
    	// assume format is correct
    	tmpp=0;
    	for(i=1; i<safestrlen(exp)-1; i++) {
    		if (exp[i]=='&') {
    			// If an AND operator
    			operator = 0;
    		} else if (exp[i]=='|') {
    			// If an OR operator
    			operator = 1;
    		} else if (exp[i]==')') {
    			// Closing bracket, end expression and evaluate
    			
    			// Finish string
    			tmp[tmpp] = ')';
    			tmp[tmpp+1] = '\0';
    			
    			// Evaluate
    			{ void, void, token_ptr }  = expression_gen(from_services, token_ptr, tmp);
    			
    			// Do AND or OR
    			if (operator==0) {
    				// AND
    				token_ptr = add_token(from_services, MTOK_BOOLEXP, 0, IS_SYSVAR, SYSVAR_RESULT, IS_SYSVAR, AND, token_ptr);
    			} else {
    				// OR
    				token_ptr = add_token(from_services, MTOK_BOOLEXP, 0, IS_SYSVAR, SYSVAR_RESULT, IS_SYSVAR, OR, token_ptr);
    			}
    			
    			// Move to 0
    			token_ptr = add_token(from_services, MTOK_SET, 0, IS_SYSVAR, SYSVAR_RESULT, IS_SYSVAR, ASSIGN, token_ptr);
    			
    			// Reset stuff
    			tmpp = 0;
    		} else {
    			// Do copy
    			tmp[tmpp] = exp[i];
    			tmpp++;
    		}
    	}
    	
    	// Final move from zero to result
    	token_ptr = add_token(from_services, MTOK_SET, SYSVAR_RESULT, IS_SYSVAR, 0, IS_SYSVAR, ASSIGN, token_ptr);
    	
    	return token_ptr;
    }
    
    return token_ptr;
}

// The expression generation function generates all the tokens required to perform an
// expression. It then returns the correct values for the val1, val2, type1, type2, op for
// the final instruction that uses the result of the expression.

// At the end of expression evaluation the expression result is stored either directly, in a var
// or in the IN_SYSVAR (result 2) store. It also returns an updated token_ptr.
{ int, use_type, int } expression_gen(chanend from_services, int token_ptr, char param[]) {
	int v1, v2;
	use_type u1, u2;
	op_type op;
	
	// Pass to split op
	{ v1, u1, op, v2, u2, token_ptr }  = split_op_argument(from_services, param, token_ptr);

	// We need to generate instructions
	if (op==-1) {
		// Individual int or var
		return { v1, u1, token_ptr };
	} else if ((op==PLUS)||(op==MINUS)||(op==MULTIPLY)||(op==DIVIDE)) {
		// This does the arithmetic operation
		token_ptr = add_token(from_services, MTOK_ARITH, v1, u1, v2, u2, op, token_ptr);
	} else {
		// And this one logical
		token_ptr = add_token(from_services, MTOK_BOOLEXP, v1, u1, v2, u2, op, token_ptr);
	}
	
	return { SYSVAR_RESULT, IS_SYSVAR, token_ptr };
}

// Map from "special ids" to a mask to use on the register
// If is not a special id, return 0.
int special_id_to_int(chanend from_services, char name[]) {
	
	// Check against text
	if (safestrcmp(name, "button_a")==0) {
		return BUTTON_A;
	} else if (safestrcmp(name, "button_b")==0) {
		return BUTTON_B;
	} else if (safestrcmp(name, "button_c")==0) {
		return BUTTON_C;
	} else if (safestrcmp(name, "button_jc")==0) {
		return BUTTON_JC;
	} else if (safestrcmp(name, "joy_up")==0) {
		return JOY_UP;
	} else if (safestrcmp(name, "joy_down")==0) {
		return JOY_DOWN;
	} else if (safestrcmp(name, "joy_left")==0) {
		return JOY_LEFT;
	} else if (safestrcmp(name, "joy_right")==0) {
		return JOY_RIGHT;
	} else if (safestrcmp(name, "key_enter")==0) {
		return KEY_ENTER;
	}else if (safestrcmp(name, "key_space")==0) {
		return KEY_SPACE;
	} else if (safestrcmp(name, "key_right")==0) {
		return KEY_RIGHT;
	} else if (safestrcmp(name, "key_left")==0) {
		return KEY_LEFT;
	} else if (safestrcmp(name, "key_down")==0) {
		return KEY_DOWN;
	} else if (safestrcmp(name, "key_up")==0) {
		return KEY_UP;
	}else {
		return 0;
	}
	
}

// An easy way to add a token
// Return token ptr
int add_token(chanend from_services, machine_tok tok_name, int v1, use_type u1, int v2, use_type u2, int op, int token_ptr) {
	// Boundary check
	if (token_ptr>=MAX_TOKENS) {
		services_graphics_add_text(from_services, "Error: Program is too big.");
		services_graphics_new_line(from_services);
		error=1;
		return token_ptr;
	}
	
	// Add token
	toks[token_ptr].type = tok_name;
	toks[token_ptr].value = v1;
	toks[token_ptr].val_type = u1;
	toks[token_ptr].value2 = v2;
	toks[token_ptr].val2_type = u2;
	toks[token_ptr].operation = op;
	
	return ++token_ptr;
}

// Split an argument of the type x=y into its components
// where x and y could be variable names or integers
{ int, use_type, op_type, int, use_type, int } split_op_argument(chanend from_services, char param[], int token_ptr) {
	// The most important thing is to detect:
	// - Are they variables or integers?
	// - What is the operator?
	// - If the variables haven't been mapped, return a warning
	
	// We mandate that:
	// - Variables must be [a-z][z-a|0-9]*
	// - Integers must be (-)?[0-9]+
	// - A single command operator is a mathematical or assignment operand
	// - A double command operator is a boolean operator
	
	char tmpb[MAX_TOKEN_LEN+1];
	int tmpbp=0;
	
	int i=0;
	
	int val1, val2;
	use_type use1, use2;
	op_type oper;

	// Check for leading bracket
	if (param[i]=='(') i++;
	
	// Check first character
	if ((param[i]=='-') || ((param[i]>=0x30) && (param[i]<=0x39))) {
		// Is int
		use1 = IS_INTEGER;
	} else if ((param[i]>=0x61) && (param[i]<=0x7A)) {
		// Is varname
		use1 = IS_VARIABLE;
	} else {
		// Is error!
		services_graphics_add_text(from_services, "Error: Unknown variable or integer. Format is not correct.");
		services_graphics_new_line(from_services);
		error=1;
	}
	
	// Read characters in until hits an operator
	for (; i<safestrlen(param); i++) {
		// Break if op
		if (is_operator(param[i])) break;
		
		// Copy characters to tmpb
		tmpb[tmpbp] = param[i];
		tmpbp++;
	}
	
	// Check
	if (use1==IS_INTEGER) {
		// Parse int and store
		tmpb[tmpbp] = '\0';
		
		val1 = string2int(tmpb);
	} else {
		int mask;
		
		// Parse var name and store
		tmpb[tmpbp] = '\0';

		// Double check the variables
		// Is either a var or a "special id" such as BUTTON_A. If it's a var,
		// we convert to varid and return, otherwise we insert an arithmetic
		// operation which stores the value of the special id into the SYSRES
		// register.

		if ((mask = special_id_to_int(from_services, tmpb))!=0) {
			// Is special id
			
			// Generate an arithemtic operation to get an AND
			token_ptr = add_token(from_services, MTOK_ARITH, SYSVAR_INPUT, IS_SYSVAR, mask, IS_INTEGER, BITWISE_AND, token_ptr);
						
			// Result is in SYSRES
			val1 = SYSVAR_RESULT;
			use1 = IS_SYSVAR;
		} else {
			// Store as var
			val1 = string2varid(tmpb);
		}
		
		// Error
		if (val1<0) {
			services_graphics_add_text(from_services, "Warning: Out of variable storage space for ");
			services_graphics_add_text(from_services, tmpb);
			services_graphics_new_line(from_services);
			error=1;
		}
	}
	
	// Check if its gone over the limit meaning no more portions
	if (i>=safestrlen(param)) {
		return { val1, use1, -1, -1, -1, token_ptr };
	}
	
	// Check operator
	if (is_operator(param[i])) {
		if ((is_operator(param[i + 1])) && (param[i+1]!='-')) {
			// Unfortunately - has tokubetsuna imi

			// In a two-digit operator the second is always an equals
			if (param[i]=='=') {
				oper = EQUAL;
			} else if (param[i]=='!') {
				oper = NOTEQUAL;
			} else if (param[i]=='<') {
				oper = LESSEQUAL;
			} else if (param[i]=='>') {
				oper = GREATEREQUAL;
			}
			
			// Two digit operator
			i++;
		} else {
			// One digit operator

			if (param[i]=='=') {
				oper = ASSIGN;
			} else if (param[i]=='*') {
				oper = MULTIPLY;
			} else if (param[i]=='/') {
				oper = DIVIDE;
			} else if (param[i]=='+') {
				oper = PLUS;
			} else if (param[i]=='-') {
				oper = MINUS;
			} else if (param[i]=='<') {
				oper = LESSTHAN;
			} else if (param[i]=='>') {
				oper = GREATERTHAN;
			} else if (param[i]=='@') {
				oper = AT;
			}
		}
	}

	// Inc
	i++;
	
	// Check first character
	if ((param[i]=='-') || ((param[i]>=0x30) && (param[i]<=0x39))) {
		// Is int
		use2 = IS_INTEGER;
	} else if ((param[i]>=0x61) && (param[i]<=0x7A)) {
		// Is varname
		use2 = IS_VARIABLE;
	} else {
		// Is error!
		services_graphics_add_text(from_services, "Error: Unknown variable or integer. Format is not correct.");
		services_graphics_new_line(from_services);
		error=1;
	}
	
	// Read characters in until hits end
	tmpbp=0;
	for (; i<safestrlen(param); i++) {
		// Check for ending backet
		if (param[i]==')') continue;
		
		// Copy characters to tmpb
		tmpb[tmpbp] = param[i];
		tmpbp++;
	}
	
	// Check
	if (use2==IS_INTEGER) {
		// Parse int and store
		tmpb[tmpbp] = '\0';

		val2 = string2int(tmpb);
	} else {
		int keyid;
		
		// Parse var name and store
		tmpb[tmpbp] = '\0';
		
		// Check it's not a special id
		if ((keyid = special_id_to_int(from_services, tmpb))!=0) {
			// Set value and use
			val2 = keyid;
			use2 = IS_INTEGER;
		} else {
			// Get varid
			val2 = string2varid(tmpb);
	
			// Error
			if (val2<0) {
				services_graphics_add_text(from_services, "Warning: Out of variable storage space for ");
				services_graphics_add_text(from_services, tmpb);
				services_graphics_new_line(from_services);
				error=1;
			}
		}
	}
	
	// Finally, return
	return { val1, use1, oper, val2, use2, token_ptr };
}

// If a var has been stored in the variable database, return its ID
// Otherwise store in the db and return the new ID
// If out of space, return -1
int string2varid(char name[]) {
	int i;
	
	// See if in database
	for (i=0; i<max_vars; i++) {
		if (safestrcmp(name, store[i].name)==0) {
			return i;
		}
	}
	
	// Not in DB! Allocate and GO!
	if (max_vars+1>NO_OF_VARS) {
		return -1;
	} else {
		// Store
		safestrcpy(store[max_vars].name, name);
		store[max_vars].value = 0;
		
		// Inc
		max_vars++;
		
		return max_vars-1;
	}
}

// Is it an operator?
int is_operator(char i) {
	if ((i=='=')||(i=='!')||(i=='<')||(i=='>')||(i=='*')||(i=='/')||(i=='+')||(i=='-') || (i=='@')) {
		return 1;
	}
	
	return 0;
}

// Remove any excess whitespace
// Combine all spaces into one and remove eg tabs
// Set all uppercase letters to lowercase
void remove_whitespace(char command[]) {
	int i,j;
	int prev_ws=0;
	int this_ws=0;
	int found_char=0;
	
	
	// Convert all whitespace chars to a space
	for (i=0; i<safestrlen(command); i++) {
		if ((command[i]=='\n') || (command[i]=='\r') || (command[i]=='\t')) command[i]=' ';
	}
	
	// Now remove excess
	for (i=0; i<safestrlen(command);) {
		// Put uppercase letters into lowercase
		if ((command[i]>=0x41) && (command[i]<=0x5A)) {
			command[i] += 32;
		}

		// Mark as found char
		if ((command[i]>=0x61) && (command[i]<=0x7A)) {
			found_char=1;
		}

		// Set
		if (command[i]==' ') this_ws=1;
		
		// Remove all tabs and crs, remove duplicate spaces and lines
		if (((this_ws) && (prev_ws)) || ((this_ws) && ((found_char==0)||(command[i+1]=='\0')))) {
			// Remove
			for (j=i; j<safestrlen(command); j++) {
				command[j] = command[j+1];
			}
		} else {
			i++;
		}
		
		// Reset
		prev_ws = this_ws;
		this_ws = 0;
	}
	
	// Can leave the end space on the end; check for it and remove it
	if ((safestrlen(command)>1) && (command[safestrlen(command)-1]==' ')) command[safestrlen(command)-1]='\0';
}
