// Simple keyboard handler

// Import header
#include "xgc_keyboard.h"
#include "xgc_kbd_decode.h"

// Global vars
#define BUFF_MAX 16
static int buffer[BUFF_MAX+1];
static unsigned buff_start=0;		// Read at start
static unsigned buff_end=0;		// Write at end
static int has_req=0;

// Define ports
port ps2_clock = XS1_PORT_1I;
port ps2_data = XS1_PORT_1L;

// Validate parity, adapted from Roger's code on X(MOS)Linkers
int validate_odd_parity(char press, int bit) {
	unsigned char intermediate = press;
	unsigned char bitc = (char)bit;
	
	intermediate = (intermediate >> 4) ^ intermediate;
	intermediate = (intermediate >> 2) ^ intermediate;
	intermediate = ((intermediate >> 1) ^ intermediate) & 0x01 ;
    
    if (intermediate!=bitc) return 1; else return 0;
}

// Look for falling edge, wait a short period read data, look for rising edge and return
int read_bit(chanend send, streaming chanend req, port ps2_clock, port ps2_data, int timeout) {
	int not_done=1;
	unsigned temp;
	unsigned tme;
	timer t;
	
	t :> tme;
	
	while (not_done) {
		select {
			case ps2_clock when pinseq(0x0) :> void:
				not_done=0;
				break;
			case req :> int discard:
				// Send a char
				if (buff_start!=buff_end) {
					send <: buffer[buff_start];
					if (buff_start==BUFF_MAX) buff_start=0; else buff_start++;
					has_req = 0;
				} else {
					has_req = 1;
				}
				break;
			case timeout => t when timerafter(tme+200000) :> void:
				return -1;
		}
	}
	
	ps2_data :> temp;
	
	select {
		case ps2_clock when pinseq(0x1) :> void:
			break;
		case t when timerafter(tme+200000) :> void:
			return -1;
	}

	return temp;
}

// Main function
void keyboard(chanend chars, streaming chanend req) {
	// Bits for control
	unsigned int start;
	unsigned int parity;
	unsigned int stop;
	
	// Press will contain the scancode, char the decoded character
	unsigned int press;
	unsigned int press_char;
	
	// This process will maintain the state of shift and control
	unsigned int shifton=0;	// Shift on?
	unsigned int ctrlon=0; // Ctrl on?
	unsigned int released=0; // Was previous char a release char?
	unsigned int extended=0; // Was previous char the extended keyboard character E0?
	unsigned int capslock=0;

	// Loop
	while (1) {
		// Reset
		press=0;
		press_char=0;
		
		// Find start bit
		while ((start = read_bit(chars, req, ps2_clock, ps2_data, 0))!=0);
		
		// Next 8 bits are the actual key
		for (int i=0; i<8; i++) {
			// Read bit into temp
			int tmp = read_bit(chars, req, ps2_clock, ps2_data, 1);
			press_char >>= 1;
			if (tmp) press_char = press_char | 0x80;
		}

		// Next is parity and stop
		parity = read_bit(chars, req, ps2_clock, ps2_data, 1);
		stop = read_bit(chars, req, ps2_clock, ps2_data, 1);

		// Check start bit
		if (start) printstr("wrong start");
		
		// Verify parity bit
		if(!validate_odd_parity(press_char, parity)) printstr("wrong parity");
		
		// Check stop bit
		if (!stop) printstr("wrong stop");

		// Decode key
		if (shifton) {
			press_char = decode_shift_press(press_char);
		} else {
			press_char = decode_press(press_char);
		}
		
		// Is this char a release?
		if (press_char==PS2_RELEASE) {
			released=1;
			extended=0;
		} else {
			if (press_char==PS2_SHIFT) {
				// Is this char a shift?
				if (released==1) shifton=0; else shifton=1;
				// Not released!
				released=0;
			} else if (press_char==PS2_CTRL) {
				// Is this char a control?
				if (released==1) ctrlon=0; else ctrlon=1;
				// Not released!
				released=0;
			} else if (press_char==PS2_EXT) {
				// Log that the next one is an extended key
				// Released again to ignore 2 key presses
				extended = 1;
			} else if (press_char==PS2_CAPS){
				// Set
				capslock++;
				
				// Check
				if (capslock==4) capslock=0;
			}else {
				if ((released!=1) && (press_char!=PS2_NOTAKEY)) {
					
					// If extended, set the topmost bit
					if (extended) {
						press_char = press_char | 0x80000000;
					}
					
					// If control is on, set upper bits
					if (ctrlon) {
						press_char = press_char | 0x40000000;
					}

					// If caps
					if (capslock==2) {
						if ((press_char>=0x61) && (press_char<=0x7A)) {
							press_char -= 32;
						}
					}
					
					// Store in buffer
					buffer[buff_end] = press_char;
					if (buff_end==BUFF_MAX) buff_end = 0; else buff_end++;
				
					// Check
					if ((has_req) && (buff_start!=buff_end)) {
						chars <: buffer[buff_start];
						if (buff_start==BUFF_MAX) buff_start=0; else buff_start++;
						has_req = 0;
					}
				}
				// Not released!
				released=0;
			}
		}
	}
}
