
// SD card functions

#include "xgc_sdcard.h"
#include <safestring.h>

// Is an sd card inserted? 0 success
int sdcard_check(chanend sdc) {
	if (FAT16_Clnt_initialise(sdc) == XMOS_SUCCESS) return 0; else return 1;
}

// Try to read from sdc
int sdcard_load(chanend sdc, char filename[], char dest[], int dest_size) {
	FP_t filep[1];
	char tmp[500];
	int rtn=1;
	int destp = 0;
	int i=0;

	if (safestrlen(filename)>=20) return 4;
	
	// Check card
	if (sdcard_check(sdc)==0) {
		// Open file
		if ((FAT16_Clnt_fopen(sdc, filep, filename, 'r')) == XMOS_SUCCESS) {
			
			// Read file
			while (rtn!=0) {
				// Get next 500 bytes
				rtn = FAT16_Clnt_fread(sdc, filep, tmp, 1, 500);

				// If destp + rtn is > dest_size+1 then finish
				if (destp+rtn>=(dest_size-1)) break;
				
				// Copy to dest
				for (i=0; i<rtn; i++) {
					dest[destp] = tmp[i];
					destp++;
				}
			}
			
			// Close
			FAT16_Clnt_fclose(sdc, filep);
		
			if (destp!=0) {
				dest[destp] = '\0';
				
				return 0;
			} else {
				return 3;
			}
		} else {
			return 2;
		}
	} else {
		return 1;
	}
}

// Write data to sdcard, don't include \0 in src_size
int sdcard_write(chanend sdc, char filename[], char src[], int src_size) {
	FP_t filep[1];
	// Temp buffer
	char tmp[500];
	int tmpp = 0;
	int overallp = 0;
	int rtn=0;
	
	// Check filename
	if (safestrlen(filename)>MAX_FILENAME) return 4;
	
	// Try to connect
	if (sdcard_check(sdc)!=0) return 1;
	
	// Write to SDC
	if ((rtn = FAT16_Clnt_fopen(sdc, filep, filename, 'w')) != XMOS_SUCCESS) { return 2; }

	while (overallp<src_size) {
		// Copy char
		tmp[tmpp] = src[overallp];
		
		// Inc
		tmpp++;
		overallp++;
		
		// If we are at 500b send
		if (tmpp==500) {
			
			// Since arrays are indexed from 0, write tmpp
			rtn = FAT16_Clnt_fwrite(sdc, filep, tmp, 1, tmpp);

			// If no bytes written is incorrect, problem
			if (rtn!=tmpp) return 1;
			
			// Written 500, overallp stays pointing to src and we reset tmpp
			tmpp=0;
		}
	}
	
	// Mop-up
	if (tmpp!=0) {
		// Write remainder
		rtn = FAT16_Clnt_fwrite(sdc, filep, tmp, 1, tmpp);

		// If not the same, problem
		if (rtn!=tmpp) return 1;
	}
	
	// No probs, close file
	FAT16_Clnt_fclose(sdc, filep);

	// And return
	return 0;
}

// Delete from sdc
// 0 is success!
int sdcard_delete(chanend sdc, char filename[]) {
	FP_t filep[1];
	int rtn=0;

	// Range check
	if (safestrlen(filename)>=MAX_FILENAME) return 4;
	
	// Check card
	if (sdcard_check(sdc)==0) {
			// So delete
			if ((rtn = FAT16_Clnt_rm(sdc, filep, filename)) != XMOS_SUCCESS) {
				return 2;
			}
			
			return 0;
	} else {
		return 1;
	}
}

// Does a file exit?
// Remember zero is yes!
int sdcard_exists(chanend sdc, char filename[]) {
	FP_t filep[1];

	// Range check
	if (safestrlen(filename)>=20) return 4;
	
	// Check card
	if (sdcard_check(sdc)==0) {
		// Open file
		if (FAT16_Clnt_fopen(sdc, filep, filename, 'r') == XMOS_SUCCESS) {
			FAT16_Clnt_fclose(sdc, filep);
			return 0;
		} else {
			return 2;
		}
	} else {
		return 1;
	}
}

// Return an array of directory names separated by ' '
int sdcard_cat(chanend sdc, char dest_arr[], int dest_size) {
	DIR_t dir[1];
	int dest_p=0;
	int i=0;
	
	// Check inserted
	if (sdcard_check(sdc)==0) {
		// Try to open dir
		if (FAT16_Clnt_opendir(sdc, ".", dir) == XMOS_SUCCESS) {	
			// Look until no files
			while (FAT16_Clnt_readdir(sdc, dir) != 0) {
				for (i=0; i<safestrlen(dir[0].name); i++) {
					// If full
					if (dest_p>=dest_size) {
						break;
					}
					
					// Else
					dest_arr[dest_p] = dir[0].name[i];
					dest_p++;
				}
				
				dest_arr[dest_p] = ' ';
				dest_p++;
			}
			
			dest_arr[dest_p] = '\0';
			
			return 0;
		} else {
			return 5;
		}
	} else {
		return 1;
	}
}
