// Main services kernel

// Kernel should manage access to SD card, keyboard, graphics
// Threads should be able to ask kernel to give control back to shell
// User can then resume the thread via the shell

// Include
#include "xgc_services.h"

// Private vars
// ------------------------------------------------------------------------------------------------------------

// Global store for previous buttons status
int prev_input_status=0;

// One character keyboard buffer for non-blocking I/O
int kbd_buffer=-1;

// Keyboard mode
// 0 - Blocking keyboard mode (default)
// 1 - Non-blocking keyboard mode (via AK_KEY_SETMODE)
int kbd_mode=0;
int kbd_sentreq=0;

// Port and clock for speaker
buffered out port:32 p_spk = XS1_PORT_1K;
// Clock
clock c_spk = XS1_CLKBLK_3;

// Ports for joystick and buttons
in port butt = XS1_PORT_4E;
in port joystick = XS1_PORT_8D;	// bit 4-7

// Private functions
// ------------------------------------------------------------------------------------------------------------

// Main services
void xgc_services(chanend services_kbd_chars, streaming chanend services_kbd_req, chanend services_sd, chanend services_shell) {
	// Configure the port and clock for speaker
	set_clock_ref(c_spk);						// Ref clock is 100MHz
	configure_out_port(p_spk, c_spk, 0);		// Port clocked to 100MHz
	start_clock(c_spk);							// Start

	// Now loop and provide services
	while (1) {
		int request=0;
		
		// Multiplex requests
		#pragma ordered
		select {
			case kbd_mode => services_kbd_chars :> request:
				kbd_buffer = request;
				break;
			case services_shell :> request:
				request_handler(request, services_shell, services_kbd_chars, services_kbd_req, services_sd);
				break;
			default:
				break;
		}
	}
}

// Handle different types of request
void request_handler(int request, chanend app, chanend services_kbd_chars, streaming chanend services_kbd_req, chanend services_sd) {
	if (request==AK_KEY_SETMODE)				ak_key_setmode(app);
	if (request==AK_KEY_REQ) 					ak_key_req(services_kbd_chars, services_kbd_req, app);
	if (request==AK_KEY_NBREQ) 					ak_key_nbreq(services_kbd_chars, services_kbd_req, app);
	if (request==AK_GRAPHICS_ADD_CHAR)			ak_graphics_add_char(app);
	if (request==AK_GRAPHICS_DEL_PREV_CHAR)		ak_graphics_del_prev_char();
	if (request==AK_GRAPHICS_NEW_LINE)			ak_graphics_new_line();
	if (request==AK_GRAPHICS_CS)				ak_graphics_cs();
	if (request==AK_SDCARD_READ)				ak_sdcard_read(app, services_sd);
	if (request==AK_SDCARD_WRITE)				ak_sdcard_write(app, services_sd);
	if (request==AK_SDCARD_DEL)					ak_sdcard_del(app, services_sd);
	if (request==AK_SDCARD_INSERTED)			ak_sdcard_inserted(app, services_sd);
	if (request==AK_SDCARD_EXISTS)				ak_sdcard_exists(app, services_sd);
	if (request==AK_SDCARD_CAT)					ak_sdcard_cat(app, services_sd);
	if (request==AK_GRAPHICS_SET_LINE)			ak_graphics_set_line(app);
	if (request==AK_GRAPHICS_SET_CHAR)			ak_graphics_set_char(app);
	if (request==AK_SPEAKER_TONE)				ak_speaker_tone(app);
	if (request==AK_GET_INPUT_STATUS)			ak_get_input_status(app);
}

// Set the keyboard mode
void ak_key_setmode(chanend app) {
	// Receive the mode
	int new_mode;
	app :> new_mode;
	
	// Make some changes depending on the selected mode
	if (new_mode==0) {
		// Mode changed to blocking
		kbd_mode = 0;
		
		// Clear buffer
		kbd_buffer = -1;
	} else {
		// Mode changed to non-blocking
		kbd_mode = 1;
		kbd_sentreq = 0;
	}
}

// Non-blocking key request
void ak_key_nbreq(chanend services_kbd_chars, streaming chanend services_kbd_req, chanend app) {
	// Send req
	if ((kbd_buffer==-1) && (kbd_sentreq!=1)) {
		// Send a request
		services_kbd_req <: 0;
		kbd_sentreq=1;
		
		// Send back -1
		app <: kbd_buffer;
	} else {
		//if (kbd_buffer!=-1) printcharln((char)kbd_buffer);
		// In a non-blocking request we return the buffer, or -1 if the buffer is empty
		app <: kbd_buffer;
	
		// Reset req
		if (kbd_buffer!=-1) kbd_sentreq=0;
		
		// Reset buffer
		kbd_buffer = -1;
	}
}

// Get the status of the various input devices
void ak_get_input_status(chanend app) {
	int final=0;
	int butt_status;
	int joystick_status;
	
	// Grab and process button status
	butt :> butt_status;
	
	if ((butt_status&0x1)==0) {
		final = (final|0x00000001);					// Button A
	}
			
	if ((butt_status&0x2)==0) {
		final = (final|0x00000004);					// Button C
	}
	
	if ((butt_status&0x4)==0) {
		final = (final|0x00000002);					// Button B
	}
	
	if ((butt_status&0x8)==0) {
		final = (final|0x00000008);					// Button JC
	}

	joystick :> joystick_status;

	if ((joystick_status&0x10)==0) {
		final = (final|0x00000040);					// Left
		//printstrln("l");
	}
	
	if ((joystick_status&0x20)==0) {
		final = (final|0x00000080);					// Right
		//printstrln("r");
	}
	
	if ((joystick_status&0x40)==0) {
		final = (final|0x00000010);					// Up
		//printstrln("u");
	}
	
	if ((joystick_status&0x80)==0) {
		final = (final|0x00000020);					// Down
		//printstrln("d");
	}
	
	// Send the final string
	app <: final;
}

// Send a key to the application
void ak_key_req(chanend services_kbd_chars, streaming chanend services_kbd_req, chanend app) {
	int key=0;
	
	// Attempt to get a key
	services_kbd_req <: 0;
		
	// Receive
	services_kbd_chars :> key;
	
	// When we have a key send a reply
	app <: key;
}

// Print a char to the screen
void ak_graphics_add_char(chanend app) {
	char ch;
	
	// Receive the char
	app :> ch;
	
	// Call the graphics library
	graphics_add_char(ch);
}

// Delete previous char
void ak_graphics_del_prev_char() {
	graphics_del_prev_char();
}

// Add new line
void ak_graphics_new_line() {
	graphics_new_line();
}

// Clear screen
void ak_graphics_cs() {
	graphics_clear_screen();
}

// Set screen position
void ak_graphics_set_line(chanend app) {
	int pos;
	
	// Receive the new position
	app :> pos;
	
	// Call the graphics library
	graphics_cursor_set_line(pos);
}
void ak_graphics_set_char(chanend app) {
	int pos;
	
	// Receive the new position
	app :> pos;
	
	// Call the graphics library
	graphics_cursor_set_char(pos);
}

// SD Card read
void ak_sdcard_read(chanend app, chanend services_sd) {
	// Filename
	char filename[21];
	int filenamesize=0;
	// Data
	char dest[MAX_SD_READ+1];
	int destsize=0;
	// Return/req status
	int rtn=0;
	// Counter
	int i=0;
	
	// Receive filename size
	app :> filenamesize;

	slave {
		for (i=0; i<filenamesize; i++) {
			app :> filename[i];
		}
	}
	
	// Call the function
	rtn = sdcard_load(services_sd, filename, dest, MAX_SD_READ);
	
	// Send status and data
	if (rtn==0) {
		// Get size
		destsize = safestrlen(dest)+1;
		
		// OK
		app <: KA_SDCARD_OKAY;
		app <: destsize;

		// Now send data
		master {
			for (i=0; i<destsize; i++) {
				app <: (char) dest[i];
			}
		}
	} else {
		// Problem
		app <: KA_SDCARD_ERROR;
		app <: rtn;
	}
}

// SD card write
void ak_sdcard_write(chanend app, chanend services_sd) {
	// Filename
	char filename[21];
	int filename_size=0;
	// Data
	char src[MAX_SD_READ+1];
	int src_size=0;
	// Return/req status
	int rtn=0;
	// Counter
	int i=0;
	
	// Size of filename plus null
	app :> filename_size;
	
	// Recieve filename
	slave {
		for (i=0; i<filename_size; i++) {
			app :> filename[i];
		}
	}
	
	// Now send the data size
	app :> src_size;

	// And now the data
	slave {
		for (i=0; i<src_size; i++) {
			app :> src[i];
		}
	}
	
	// Call function
	rtn = sdcard_write(services_sd, filename, src, src_size);
	
	// Return
	if (rtn==0) {
		app <: KA_SDCARD_OKAY;
	} else {
		app <: KA_SDCARD_ERROR;
		app <: rtn;
	}
}

// SD Card del
void ak_sdcard_del(chanend app, chanend services_sd) {
	// Filename
	char filename[21];
	int filename_size=0;
	// Counter
	int i;
	// Return
	int rtn=0;
	
	// Size of filename plus null
	app :> filename_size;
	
	// Receive filename
	slave {
		for (i=0; i<filename_size; i++) {
			app :> filename[i];
		}
	}
	
	// Call function
	rtn = sdcard_delete(services_sd, filename);
	
	// Send status status
	if (rtn==0) {
		app <: KA_SDCARD_OKAY;
	} else {
		// Send error
		app <: KA_SDCARD_ERROR;
		// Send error code
		app <: rtn;
	}
}

// SDC inserted?
void ak_sdcard_inserted(chanend app, chanend services_sd) {
	int rtn=0;
	
	// Call func and return
	rtn = sdcard_check(services_sd);

	app <: rtn;
}

// Does a file exist?
void ak_sdcard_exists(chanend app, chanend services_sd) {
	// Filename
	char filename[21];
	int filename_size=0;
	// Counter
	int i;
	// Return
	int rtn=0;

	// Size of filename plus null
	app :> filename_size;
		
	// Send filename
	slave {
		for (i=0; i<filename_size; i++) {
			app :> filename[i];
		}
	}
	
	// Function
	rtn = sdcard_exists(services_sd, filename);
	
	// Send
	app <: rtn;
}

// Catalogue the SDC
void ak_sdcard_cat(chanend app, chanend services_sd) {
	char cat[MAX_SD_READ+1];
	int cat_len=0;
	
	int rtn=0;
	int i=0;

	// Run func
	rtn = sdcard_cat(services_sd, cat, MAX_SD_READ);
	
	// If return is non-zero, return the errorcode, else
	// OK and then data
	if (rtn==0) {
		// OK
		app <: KA_SDCARD_OKAY;
		
		// Len
		cat_len = safestrlen(cat)+1;
		app <: cat_len;
		
		// Return the data
		master {
			for (i=0; i<cat_len; i++) {
				app <: (char) cat[i];
			}
		}
		
	} else {
		// Not OK
		app <: KA_SDCARD_ERROR;
		app <: rtn;
	}
}

// Public functions
// ------------------------------------------------------------------------------------------------------------

// Print a char to the screen
void services_graphics_add_char(chanend from_services, char ch) {			
	from_services <: AK_GRAPHICS_ADD_CHAR;
	from_services <: (char) ch;
}

// New line
void services_graphics_new_line(chanend from_services) {
	from_services <: AK_GRAPHICS_NEW_LINE;
}

// Print an array to screen
void services_graphics_add_text(chanend from_services, char array[]) {
	 int i=0;
	 for (i=0; i<safestrlen(array); i ++) {
		 from_services <: AK_GRAPHICS_ADD_CHAR;
		 from_services <: (char) array[i];
	 }
}

// Clear the screen
void services_graphics_cs(chanend from_services) {
	from_services <: AK_GRAPHICS_CS;
}

// Set line and pos
void services_graphics_set_line(chanend from_services, int pos) {
	from_services <: AK_GRAPHICS_SET_LINE;
	from_services <: pos;
}
void services_graphics_set_char(chanend from_services, int pos) {
	from_services <: AK_GRAPHICS_SET_CHAR;
	from_services <: pos;
}

// Load a file from SD Card
int services_sdcard_load(chanend from_services, char filename[], char dest[], int dest_size) {
	// A request
	int req=0;
	// Data size temp
	int data_size=safestrlen(filename)+1;
	// Counter
	int i;
	
	// Send an SD Card request
	from_services <: AK_SDCARD_READ;
	from_services <: data_size;
	
	// Send filename
	master {
		for (i=0; i<data_size; i++) {
			from_services <: (char) filename[i];
		}
	}
	
	// Receive data or error code
	from_services :> req;
	
	if (req==KA_SDCARD_OKAY) {
		// What size is the received data?
		from_services :> data_size;
		
		// Just check
		if (data_size>dest_size) return 4;
		
		
		// Start stream
		slave {
			for(i=0; i<data_size; i++) {
				from_services :> dest[i];
			}
		}
			
		// Return
		return 0;
	} else {
		// KA_SDCARD_ERROR
		
		// Receive error code
		from_services :> req;
		
		// Return
		return req;
	}
}

// Write a file to the SD Card
int services_sdcard_write(chanend from_services, char filename[], char src[], int src_size) {
	// A request
	int req=0;
	// Data size temp
	int data_size=safestrlen(filename)+1;
	// Counter
	int i;
	
	// Send an SD Card request
	from_services <: AK_SDCARD_WRITE;
	
	// Size of filename plus null
	from_services <: data_size;
	
	// Send filename
	master {
		for (i=0; i<data_size; i++) {
			from_services <: (char) filename[i];
		}
	}
	
	// Now send the data size
	from_services <: src_size;

	// And now the data
	master {
		for (i=0; i<src_size; i++) {
			from_services <: (char) src[i];
		}
	}
	
	// Get the status code
	from_services :> req;
	
	// Return
	if (req==KA_SDCARD_OKAY) {
		return 0;
	} else {
		// Get error
		from_services :> req;
		return req;
	}
}

// Delete a file from the SD Card
int services_sdcard_del(chanend from_services, char filename[]) {
	// A request
	int req=0;
	// Data size temp
	int data_size=safestrlen(filename)+1;
	// Counter
	int i;
	
	// Send an SD Card request
	from_services <: AK_SDCARD_DEL;
	
	// Size of filename plus null
	from_services <: data_size;
	
	// Send filename
	master {
		for (i=0; i<data_size; i++) {
			from_services <: (char) filename[i];
		}
	}
	
	// Get status
	from_services :> req;
	
	if (req==KA_SDCARD_OKAY) {
		return 0;
	} else {
		// Get error code
		from_services :> req;
		// Return
		return req;
	}
}

// Is an SD card inserted?
int services_sdcard_inserted(chanend from_services) {
	int req=0;
	
	// Send signal
	from_services <: AK_SDCARD_INSERTED;
	
	// Get result
	from_services :> req;
	
	// Return
	return req;
}

// Does a file exist?
int services_sdcard_exists(chanend from_services, char filename[]) {
	// A request
	int req=0;
	// Data size temp
	int data_size=safestrlen(filename)+1;
	// Counter
	int i;
	
	// Send an SD Card request
	from_services <: AK_SDCARD_EXISTS;
	
	// Size of filename plus null
	from_services <: data_size;
	
	// Send filename
	master {
		for (i=0; i<data_size; i++) {
			from_services <: (char) filename[i];
		}
	}
	
	// Get status
	from_services :> req;
	
	return req;
}

// Fill an array with the cat of the sdc
int services_sdcard_cat(chanend from_services, char cat[], int cat_len) {
	int req=0;
	int actual_len;
	int i;
	
	// Send command
	from_services <: AK_SDCARD_CAT;
	
	// What happened?
	from_services :> req;
	
	if (req==KA_SDCARD_OKAY) {
		// Get len
		from_services :> actual_len;

		// Obtain the data
		slave {
			for (i=0; i<actual_len; i++) {
				if (i>cat_len) {
					// Discard
					from_services :> req;
				} else {
					from_services :> cat[i];
				}
			}
		}
		
		return 0;
	} else {
		// Not OK
		from_services :> req;
		
		return req;
	}
}

// Output to speaker
void ak_speaker_tone(chanend app) {
	// Receive the length (1/10 sec) and freq in units of 150Hz
	int len, freq, i, j, h, block;
	
	//  Generate an array to reduce the time in loop
	unsigned int outputs[5] = {
		0x55555555,
		0xEEEEEEEE,
		0xFFFFFFFF,
		0x11111111,
		0x00000000,
	};
	
	// Receive
	app :> len;
	app :> freq;

	// If we have an output of 100MHz, and a complete period is 8 values then we can for the moment
	// Human hearing range is 16Hz-16kHz.
	
	// Aiming for 1kHz actual freq => 8kHz per cycle => 12500 10ns units => 8 * 12500
	
	// For each level we need to put on values such that we get the correct amount.
	// 0 0%			Half 1s, Half 0s
	// 1 70%		
	// 2 100%		All 1s
	// 3 70%		
	// 4 0%			Half 1s, Half 0s
	// 5 -70%		
	// 6 -100%		All 0s
	// 7 -70%		
	
	// Set the number of repititions in a block
	block = ((100000000/freq)/8)/32;
	
	// Let's go
	for (h=0; h<(len*freq)/10; h++) {
		
		for (i=0; i<8; i++) {
			
			for (j=0; j<block; j++) {
				
				// Switch
				switch (i) {
					// 0%
					case 0:
					case 4:
						// Half 1s, Half 0s
						p_spk <: outputs[0];
						break;
					
					// 70%
					case 1:
					case 3:
						// 3/4 1s, 1/4 0s
						p_spk <: outputs[1];
						break;
						
					// 100%
					case 2:
						// Always 1
						p_spk <: outputs[2];
						break;
						
					// -70%
					case 5:
					case 7:
						// 3/4 0s, 1/4 1s
						p_spk <: outputs[3];
						break;
						
					// -100%
					case 6:
						// All 0s
						p_spk <: outputs[4];
						break;
				}
				
			}
			
		}
		
	}
	
}

// Utiltiies
// -----------------------------------------------------------------------------------------------------

// Substr
void substr(char src[], char dest[], int start, int end, int stripspace) {
	int destp=0;
	int i=0;

	// Extract the dest string
	for (i=start; i<end; i++) {

		if (src[i]=='\0') break;
		
		dest[destp] = src[i];
		destp++;
	}
	
	dest[destp] = '\0';
}

// K & R Implementation itoa
// From C Programming Language Second Edition

void reverse(char s[]) {
    int c, i, j;

    for (i = 0, j = safestrlen(s)-1; i<j; i++, j--) {
        c = s[i];
        s[i] = s[j];
        s[j] = c;
    }
}


void itoa(int n, char s[]) {
    int i, sign;

    if ((sign = n) < 0)  /* record sign */
        n = -n;          /* make n positive */
    i = 0;
    do {       /* generate digits in reverse order */
        s[i++] = n % 10 + '0';   /* get next digit */
    } while ((n /= 10) > 0);     /* delete it */
    if (sign < 0)
        s[i++] = '-';
    s[i] = '\0';
    reverse(s);
} 

